#!/usr/bin/env python2.3
# -*- coding: utf-8 -*-
#
#Convert the stems as stored in the single tab delimited input lexicon file
# into separate files for each POS.  This is adopted for Nahuatl from
# the generic TabDelim2MultLexc.py; the modification consists of adding
# a tab delimited field for paradigm class, and omitting the gloss field.
# There are also some special treatments of POSs, paradigm class strings
# (the lexicon uses multiple ones, we reduce this to a smaller set),
# and the form (when this contains a colon, used to mark vowel length).
# (A UID field may be added later, in place of the gloss field.)
#
# Arguments: -i <InputFile> (optional, defaults to stdin)
#            -o <OutputFile> (optional, defaults to stdout)
#            -e <ErrFile> (optional, and not currently used; defaults to stderr)
# Input: A tab-delimited file, with one record per line.  (For Nahuatl, the fields
#        are created by the makefile from the UR_SFM or LEMMA_SFM, POS_SFM, and
#        PARADIGM_SFM fields respectively.)
#        Fields:
#            (1) Target language word
#            (2) POS
#            (3) Paradigm class (may be empty)
#            (Later) UID
#        Any fields beyond the third are silently ignored.
# Output: We create separate files in the current dir in lexc format for each
#        POS found in the input that corresponds to an acceptable POS (the list
#        of acceptable POSs is hard-coded in here; we make some modifications,
#        removing parenthesized material at the end--this usually corresponds
#        to subcats, probably irrelevant to the morphology).  The separate files
#        allow , so that each POS can be loaded into separate xfst variables.
#        Each such file is named <POS>.lexc  (There is a special case for
#        Aux', because Windows/DOS makes it nearly impossible to work with
#        a file whose basename is Aux.)
#        In addition, we create single xfst file named as specified in the
#        -o arg, with the xfst cmds to load the lexc files and 'defines'
#        for each lexicon.  (But see 'To-do' below.)
#        The paradigm class is turned into a string preceded by '^' and attached
#        (without a space) to the upper (gloss) and lower (Nahuatl form) field.
#        No attempt is made to sort the entries within a given POS by paradigm class.
#
# We ignore phrasal entries (i.e. any record with a space char in the first field).
# We split multiple glosses in the input (separated by commas) into distinct
# records in the output, and replace any remaining spaces in glosses by '_'.


import sys
import getopt
import time
import string
import codecs   #For UTF-8
import re
import sets

#----------------------Constants--------------------

sNahuatlAlphabet = ( 'a',  u'á',  'b',  'c', 'd', 'e', u'é', 'f', 'g',
                     'h',   'i',  u'í', 'j', 'k', 'l',  'm',  'n', 'o',
                    u'ó',  'p',   'r',  's', 't', 'u', u'ú', 'w', 'x',
                     'y', ':'
                   )
  #If some of these chars look odd, you may need to re-open this file in a UTF-8 encoding.
  #No 'qvz' (or space, which would be used in phrasal entries, or before a comment).
  #Does the accented actually 'u' appear?
  #These are just the chars as Unicode defines them--e.g. there is not a special case
  # for 'ch' or the long vowels as units.

sVerbPOSs = ["V0", "V1", "V2", "V3", "V4"]
  #Dictionary also has e.g. "V3_refl", but we don't handle those yet.
  # Copular "verbs" (V0cop) are NOT treated as verbs here, so they don't need
  # a paradigm class (below), even though the dictionary defines one ("\infv Copular")
sInflectedCategories = sVerbPOSs + ["V0cop"]
  # "V0cop" comes from "V0(cop)", and is munged so we have a legit filename.
  # ("V0_cop" doesn't work, because xfst chokes on an underscore char.)
  #The words of these categories have paradigms in which at least one form is inflected.
  # For now, this is just verbs (including copular verbs), but other categories will be added later.
sPOSs = sInflectedCategories + \
 ["Adj", "Adj(ap)",
  "Adv", "Adv-interog", "Adv-manner", "Adv-place", "Adv-time",
  "AuxV",   #'AuxV' is a special case, because of the limitations of filenames in DOS/Windows:
            # You can't name a file 'aux'.  Other special filenames are 'CON', and maybe 'COM'.
  "Baby",
  "Call",
  "Clit",
  "Conj",
  "Coord",
  "Dem-adj",
  "Dir",
  "Discourse",
  "Funct",
  "Interj",
  "Loc",
  "Modal",
  "N", "N(ap)", "N(at)", "N(loc)", "N(proper)", "N(rel)", "N-N", "N-ad", "N-compl", "N-tsi:n", "N/Adj", "N1", "N2",
  "N; Adj", "N; Interj", #Jonathan will probably normalize these
  "Neg",
  "Noun(rel)", #Jonathan will probably normalize this
  "Num",
  "PM",
  "Pr", "Pr(dem)", "Pr(demon)", "Pr(ind)", "Pr(indef)", "Pr(int)", "Pr(obj)", "Pr(obj); Pr(poss)", "Pr(pos)", "Pr/Adj(dem)", "Pr/Adj(demon)", "Pr?",
  "Pref(obj)", "Pref(pos)", "Pref(sub)", "Prefix",
  "Pronoun",
  "Quant",
  "Reltz", "Reltzer",
  "Subord",
  "Undeter"
  ]
#PSMs commented out for now (see my 2 Apr 06 email to Jonathan):
#  "Associated motion ending",
#  "Durative",
#  "Nomlz",
#  "Stat",
#  "Suf", "Suf(adj)", "Suf(loc)", "Suf-poss",
#The ff. are mostly covered under the verbal POSs:
#  "V0", "V0(cop)", "V1", "V1(refl/only)", "V2", "V2(pass)", "V2(refl/dif)", "V2(refl/only)", "V3", "V3(refl)", "V4"

#Ff. comes from the \infv field:
sVerbalParadigmClasses = ["^C1", "^C2a", "^C2b", "^C3a", "^C3b", "^C3c", "^C3d", "^C4a", "^C4b", "^C4c", "^C4d",
                          "^C4ita", "^C4maka",
                          "^Cmati", "^Cwahlaw"
                         ]
  #For verbs, we only accept paradigm classes of this set, Classes 4ita and mati
  # do not appear as such in the dictionary.  Rather, 4ita is marked
  #    \infv Irregular; see <klamoa>ita</klamoa>
  # These behave like "ordinary class 4 verbs" (IAW Jonathan's Bug Box), except for
  # the durative stem.  They all end in /ita/ or /ata/ (or accented versions of these).
  # So we convert the above \infv field to
  #     "^C4ita"
  # The lexical form is slightly different: we replace a prefix "class-" with a "^C", and
  # strip any parenthesized sub-classes (e.g. "class-4c(tia)" becomes "^C4c").  However,
  # there needs to be special treatment for one class, 4a(oni) (see below)
  # Similarly, class mati is marked
  #    \infv Irregular; see <klamoa>mati</klamoa>
  # and this is converted into its own class "^Cmati".
  # And class wahlaw is marked
  #     \infv Irregular; see <klamoa>wa:hlaw</klamoa>
  # This is converted to its own class "^Cwahlaw" (the colon is dropped because it makes
  # xfst choke).

#----------------------Functions--------------------

def OpenInputOutput():
    #Look for command line args (if any):
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'i:o:e:')
    except getopt.GetoptError:
        print 'Usage: NahTab2MultLexc.py (-i <inputfile>) (-o <outputfile>) (-e <errorfile>)'
        sys.exit(1)

    #Defaults:
    InFile = 'stdin'
    OutFile = 'stdout'
    ErrFile = 'stderr'

    #Non-defaults:
    for o in opts:
        if o[0] == '-i':
            InFile = o[1]
        elif o[0] == '-o':
            OutFile = o[1]
        elif o[0] == '-e':
            ErrFile = o[1]

    (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')
       #We don't use the UTF8Encode or UTF8Decode, but they are returned by codecs.lookup()

    #Make sure the input stream exists:
    try:
        if InFile == 'stdin':
            InStream = UTF8Reader(sys.stdin)
        else:
            InStream = UTF8Reader(open(InFile, 'r'))
    except IOError:
        print "Error: Cannot open file %s" %InFile
        sys.exit(1)

    if OutFile == 'stdout':
        OutStream = UTF8Writer(sys.stdout)
    else:
        OutStream = UTF8Writer(open(OutFile, 'w'))

    if ErrFile == 'stderr':
        ErrStream = UTF8Writer(sys.stderr)
    else:
        ErrStream = UTF8Writer(open(ErrFile, 'w'))

    return (InStream, OutStream, ErrStream)


def isNahuatlChars(sString):
    "Return True if sString consists entirely of chars in the Nahuatl alphabet, \
     False otherwise.  Used to test stems for validity."
    for cChar in sString:
        if not cChar in sNahuatlAlphabet:
            return False
    return True


#--------------------------Class Lexeme---------------------------
class Lexeme:
    """Attributes:
          sCitationForm (String)
          sUnderlyingForm (String): usually identical to sCitationForm, but not always
          POS  (Object)
          sParadigmClass (String)
       Possibly to be added later:
          UID
       All of the work is in the __init__() method, which parses an input string into
       the various fields, performs error checking, and does normalization on the fields."""

    def __init__(self, sRecordLine):
        """Create a normalized version of sRecordLine if it is valid.  Some things we test
           for are the chars in the stem, spaces in the POS or conjugation class
           (which are often comments), backslashes in the line, etc.  If sRecordLine or any
           of the fields are invalid, we raise an exception for the first error we find."""

        #Initializations:
        self.sCitationForm   = None
        self.sUnderlyingForm = None
        self.POS             = None
        self.sParadigm       = None   #A value of '' will be used for no paradigm class

        sRecordLine = sRecordLine.strip() #Get rid of any trailing whitespace

        #Check for any backslashes, which probably mean we picked up an SFM as the contents
        # of a field:
        if sRecordLine.find("\\") != -1:
            raise Lexeme.BackslashWarning, sRecordLine

        #If we get here, the tab-delimited record does not contain any backslashes.
        try:
            #First try to split it into three parts:
            (sRawStem, sRawPOS, sRawParadigmClass) = sRecordLine.split('\t')
        except:
            #If a 3-part division didn't work, try two parts (since the paradigm class is optional):
            try:
                (sRawStem, sRawPOS) = sRecordLine.split('\t')
                sRawParadigmClass = ''  #default: no paradigm class
            except:
                raise Lexeme.FieldParseWarning, sRecordLine
        #If we get here, we have the right number of fields; ensure each field is valid,
        # and normalize them:
        sRawParadigmClass = sRawParadigmClass.strip()

        #POS:
        if string.find(sRawPOS, ' ') != -1:
            #If we get here, sRawPOS contains a space char.  We could replace that with an underscore,
            # but the space probably indicates a comment, and we don't know what to do with that.
            # So we reject the record:
            raise Lexeme.POSSpaceCharWarning, sRecordLine
        self.NormalizePOS(sRawPOS)
        if self.POS == None:
            #If we get here, the POS was not in the list of legit POSs (sPOSs), even after making
            # some changes (like stripping parenthesized material); we don't want to keep this
            # lex entry, so return False:
            raise Lexeme.UnknownPOS, sRecordLine


        #Citation form of stem:
        if sRawStem == '':
            raise Lexeme.EmptyStemWarning, sRecordLine
        if not isNahuatlChars(sRawStem):
            #Some entries have '-----', we don't want those; this also catches phrasal stems,
            # which would have a space char, as well as some bad spellings.
            raise Lexeme.StemCharsWarning, sRecordLine
        else:
            #Stem is OK, add it:
            self.NormalizeStem(sRawStem)

        #Underlying form (usually this is the same as the citation form, but for at least
        # one verb sub-subclass, it is different--and we also twiddle the paradigm class,
        # hence the return value):
        sRawParadigmClass = self.SetUnderlyingForm(sRawParadigmClass)

        #Paradigm Class (there is a check for internal space chars, but it is inside
        # NormalizeParadigmClass() because under certain circumstances--"irregular"
        # verbs that behave like 'ita'--we allow the space chars).
        self.NormalizeParadigmClass(sRawParadigmClass, sRecordLine)

        #Everything checks out;
        if self.IsInflected():
            #This word is inflected (based on its POS), so add it to the list of lexemes for the POS:
            self.POS.Lexemes.append(self)
        else:
            #This word is not inflected, so add it to the list of uninflected words:
            UninflectedForms.append(self)


    #-------------------Normalization functions-----------------------
    #These are called by the __init__() function.
    def NormalizeStem(self, sRawStem):
        """Normalize the citation form stem, and set it.  Specifically, colons
           (used to mark vowel length) get escaped with a '%' character, so lexc
           doesn't complain."""
        self.sCitationForm = sRawStem.strip().replace(":", "%:")
        return


    def NormalizePOS(self, sRawPOS):
        "Normalize the POS and set it.  We require that the POS belong to one of a pre-defined \
         set of POSs (contained in the var sPOSs, which is hard-coded in this program). \
         If this is a new raw POS, we trim off any parenthesized material from the raw POS, \
         e.g. 'V1(refl)' becomes 'V1'.  If the result is in sPOSs, we add the original raw \
         POS (including the parens) plus its translation into the normalized POS to the global \
         dictionary used to quickly normalize POSs.  \
         For all legit POSs, we update the count for this sRawPOS and its normalized form, and \
         return True.  For non-legit POSs, we return False.  (Currently, we don't actually use \
         the return value, and instead rely on the value of self.POS.)"
        if sRawPOS == '':
            self.POS = None
            return False
        else:
            if not RawPos2POS.has_key(sRawPOS):
                #This is a new raw POS, normalize the form and record it:
                sModPOS = sRawPOS
                if sModPOS == "V0(cop)":
                    #Special case for copular "verbs":
                    sModPOS = "V0cop"
                else:
                    #For everything else, create a normalized form incrementally.
                    # First strip off any trailing period:
                    if sModPOS.endswith('.'):
                        #Trim trailing period:
                        sModPOS = sModPOS[:-1]
                    #...then strip off any parenthesized material:
                    sModPOS = sModPOS.strip().split("(")[0]
                    if sModPOS.lower() == 'aux':
                        #Special case to work around the problem with filename 'AUX' (see
                        # comments at sPOSs)
                        sModPOS = 'AuxV'
                    #Capitalize first letter:
                    sModPOS = sModPOS.capitalize()
                #Now see if that gave us a legit POS:
                if sModPOS in sPOSs:
                    #It is, so add it to the dictionary using the non-normalized form
                    # as the key (for fast lookup next time), and the normalized form
                    # as the value:
                    RawPos2POS[sRawPOS] = POSs[sModPOS]
                else:
                    #If we get here, this POS is not a "legitimate" one
                    self.POS = None
                    return False
            #If we get here, the POS is OK.  If the non-normalized POS was not a key in
            # the dictionary before, it is now; use the normalized form for the record value:
            self.POS = RawPos2POS[sRawPOS]
            return True


    def SetUnderlyingForm(self, sRawParadigmClass):
        """For a few verb sub-classes, the underlying form is different from the form
           in the lexicon.  The stem is adjusted here to be the underlying form.
           As a side effect, we may twiddle the paradigm class after changing
           the underlying form.  We return that possibly modified paradigm class."""
        if sRawParadigmClass == "class-4a(oni)" or sRawParadigmClass == "class-4a(pass)":
            #Verbs of class 4a(oni) have an underlying long stem-final vowel,
            # and at the same time they cannot be distinguished from the other
            # 4a verbs phonologically.  (All the 4a(oni) verbs end in /i/,
            # but so do some 4a verbs that are not 4a(oni).)  This length
            # will be removed for the lexical stem in VerbStems.xfst.
            #Verbs of class 4a(pass) also have an underlying long stem-final vowel.
            # While verbs of this subclass of 4a can be distinguished underlyingly
            # (they all end in /lo/), for consistency we treat the underlying length
            # here too, like class 4a(oni).
            self.sUnderlyingForm = self.sCitationForm + "%:"
        elif sRawParadigmClass == "class-4b(ya-x; long)":
            #Verbs of subclass 4b(ya-x) have two sub-subclasses(!); those marked
            # "long" in the lexicon have a long /i:/ vowel in the durative, i.e. /yi:x/.
            # The other stems have short /i/, i.e. /yia/ in the durative.
            # In the lexicon, the endings for class 4b verbs are:
            #     ia
            #     ya
            #     a
            #     i
            # The class 4b(ya-x) verbs seem to have only the first of these.
            # We treat this class as having underlying length, which the various
            # stem rules in VerbStems.xfst get rid of before /a/
            # (but not before the /x/ of the durative).  Can you say "kludge"?
            # Actually, there might be some truth to this analysis...
            self.sUnderlyingForm = self.sCitationForm[:-1] + "%:" + self.sCitationForm[-1:]
            #Now that we have taken account of the sub-subclass, we can adjust the stem class
            # so that the check in NormalizeParadigmClass() for space chars doesn't flag this
            # as an error (Can you say "Kludge"?):
            sRawParadigmClass = "class-4b(ya-x)"
        else:
            #Default case:
            self.sUnderlyingForm = self.sCitationForm
        return sRawParadigmClass #Possibly modified...


    def NormalizeParadigmClass(self, sRawParadigmClass, sRecordLine):
        """Normalize Paradigm class: the lexicon uses a large set, which for verbs
           we map to a much smaller pre-defined set.  (Other POSs don't have paradigm classes.)
           (Implication: the POS has to be normalized before the paradigm class.)
           In one case, we change the lexical form of the stem to be an underlying form
           based on its stem class.  ICO verbs marked '<klamoa>ita</klamoa>', we create a new
           class 4cita; similarly, for verbs marked '<klamoa>maka</klamoa>', we create a new
           class 4maka.  In (at least) one other case 'class-4b(ya-x; long)', SetUnderlyingForm()
           sets an underlying form, then changes the paradigm class to class-4b(ya-x).
           (Can you say "Kludge"?)
           The final arg sRecordLine (= the original record, in one line) is for error msgs."""
        #Figure out the paradigm class, if any:
        if sRawParadigmClass == '':
            #Empty class
            self.sParadigm = ''  #Rather than None
        else:
            #Non-empty paradigm class, normalize it.
            #See if this is a paradigm class we've seen before:
            if not ParadigmClasses.has_key(sRawParadigmClass):
                #New paradigm class; create a normalized form:
                if self.IsVerb():
                    #Actually, this shouldn't get called for non-verbs, because sRawParadigmClass
                    # should be a null string (see if above).
                    ##FIX: The negative copular verb 'xa:k' is listed as psm = V1, infv = Copular;
                    # need special case for it
                    if "<klamoa>ita</klamoa>" in sRawParadigmClass:
                        #These verbs are marked in the lexicon as being irregular;
                        # we treat them instead as if they were a subclass of
                        # class 4 verbs (see Jonathan's bug #6).  (They contain
                        # an internal space char, so we delay the test for that--
                        # which would normally mean an erroneous paradigm class--
                        # until the other branch of this if-then).
                        sNormalizedParadigmClass = "^C4ita"
                    elif "<klamoa>maka</klamoa>" in sRawParadigmClass:
                        #These verbs are marked in the lexicon as being irregular;
                        # we treat them instead as if they were a subclass of
                        # class 4a verbs.
                        sNormalizedParadigmClass = "^C4maka"
                    elif "<klamoa>mati</klamoa>" in sRawParadigmClass:
                        #These verbs are marked in the lexicon as being irregular;
                        # we treat them as a special class, 'Cmati':
                        sNormalizedParadigmClass = "^Cmati"
                    elif "<klamoa>wa:hlaw</klamoa>" in sRawParadigmClass:
                        #The verb 'wa:hlaw' is marked in the lexicon as being irregular;
                        # we treat it as a special class, 'Cwa:hlaw':
                        sNormalizedParadigmClass = "^Cwahlaw"
                    else:
                        #For everything else, ensure that there are no space chars:
                        if string.find(sRawParadigmClass, ' ') != -1:
                            #If we get here, sRawParadigmClass contains a space char.
                            # We could replace that with an underscore, but the space probably
                            # indicates a comment, and we don't know what to do with that.
                            # So we reject the record:
                            raise Lexeme.ParadigmClassSpaceCharWarning, sRecordLine
                        else:
                            #No internal space chars, so replace the leading "class-"
                            # (which always appears for verbs) with "^C",
                            # and remove any trailing parenthesized subclass:
                            sNormalizedParadigmClass = \
                                sRawParadigmClass.replace("class-", "^C", 1).split("(")[0]
                    #Then ensure that the result is in the pre-defined set of valid
                    # paradigm classes for verbs:
                    if not sNormalizedParadigmClass in sVerbalParadigmClasses:
                        #Invalid verbal paradigm class:
                        raise Lexeme.BadParadigmClass, sRecordLine
                else:
                    #Not a verb, so the class is irrelevant (only verbs have \infv fields,
                    # but copular verbs are listed in the lexicon as "\infv Copular"; we
                    # ignore that here, giving them a null paradigm class):
                    sNormalizedParadigmClass = ""
                #Now we have a normalized form for this paradigm class.  Add the raw form as key,
                # and the normalized form as value, plus the count (initialized to 1):
                ParadigmClasses[sRawParadigmClass] = RawParadigmClass(sRawParadigmClass, sNormalizedParadigmClass)
            else:
                #Not a new paradigm class, update its count:
                ParadigmClasses[sRawParadigmClass].increment()
            #If this paradigm class wasn't in the ParadigmClasses dictionary before, it is now.
            # Assign the paradigm class to this lexeme:
            self.sParadigm = ParadigmClasses[sRawParadigmClass].NormalizedForm
            #...and make sure it's recorded for this POS:
            self.POS.MaybeAddParadigmClass(self.sParadigm)
        return


    def IsVerb(self):
        "Return True if the POS is verb (one of several: V0, V1 etc.), False otherwise. \
         (We do not currently deal with things like 'V2_refl'.) \
         Signal an error if the POS is not set."
        if self.POS == None:
            #Error--it's not set yet
            raise Lexeme.POSNotSet, sRecordLine
        else:
            #We have a POS, so proceed
#            sys.stderr.write(" %s %s\n" %(self.POS.sForm, self.POS.sForm in sVerbPOSs))
            return self.POS.sForm in sVerbPOSs

    def IsInflected(self):
        "Return True if the word is inflected (based on its POs), False otherwise. \
         Signal an error if the POS is not set."
        if self.POS == None:
            #Error--it's not set yet
            raise Lexeme.POSNotSet, sRecordLine
        else:
            #We have a POS, so proceed
#            sys.stderr.write(" %s %s\n" %(self.POS.sForm, self.POS.sForm not in sInflectedCategories))
            return self.POS.sForm in sInflectedCategories


    #---------------Lexeme class errors-------------
    class BackslashWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Record contains backslash (multiple occurences of a field?):\n\t%s\n" %sRecordLine)

    class FieldParseWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Record cannot be parsed into fields:\n\t%s\n" %sRecordLine)

    class EmptyStemWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Rejecting record with empty stem:\n\t%s\n" %sRecordLine)

    class StemCharsWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Rejecting record with invalid chars in stem:\n\t%s\n" %sRecordLine)

    class POSSpaceCharWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Rejecting record with space char in POS:\n\t%s\n" %sRecordLine)

    class UnknownPOS(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Rejecting record with unknown POS:\n\t%s\n" %sRecordLine)

    class ParadigmClassSpaceCharWarning(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Rejecting record with space char in paradigm class:\n\t%s\n" %sRecordLine)

    class POSNotSet(Exception):
        #This shouldn't happen, since we control when the POS is set (it has to be set before
        # the paradigm class can be set).
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: POS not set for this record:\n\t%s\n" %sRecordLine)

    class BadParadigmClass(Exception):
        def __init__(self, sRecordLine):
            ErrOut.write("Warning: Invalid paradigm class for this record:\n\t%s\n" %sRecordLine)


#--------------------------Class POS---------------------------
class POS:
    "Keeps track of a (normalized) POS, its lexemes, and the paradigm classes it uses."

    def __init__(self, sForm):
        #Done at program initialization.
        self.sForm = sForm
        self.Lexemes = []
        self.ParadigmClasses = []

    def MaybeAddParadigmClass(self, sParadigmClass):
        "If sParadigmClass isn't already there, add it."
        if not sParadigmClass in self.ParadigmClasses:
            self.ParadigmClasses.append(sParadigmClass)

    def WriteXFSTLoaderCode(self, XFSTStream):
        "Write the load information for this POS to the xfst file."
        XFSTStream.write('read lexc %s.lexc\ndefine %s\n\n' %(self.sForm, self.sForm))

    def WriteLexcFile(self):
        "Write out a lexc file containing the lex entries belonging to this POS."
        (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')

        LexiconOut = UTF8Writer(open(self.sForm + '.lexc', 'w'))
            #The POSs have a final '.'
        #Header for this lexc file:
        LexiconOut.write('!Generated from %s on %s by NahTab2MultLexc.py.\n\n'
                         %(LexiconIn.name, time.strftime('%d %B %Y at time %H:%M', time.localtime()))
                        )
        #Multichar symbols for any paradigm classes belonging to this POS:
        if len(self.ParadigmClasses) > 0:
            LexiconOut.write('Multichar_Symbols\n')
            for sParadigmClass in self.ParadigmClasses:
                LexiconOut.write('\t%s\n' %sParadigmClass)
            LexiconOut.write('\n')
        #...lexicon headers:
        LexiconOut.write('LEXICON Root\n\t' + self.sForm + ' ;\n\nLEXICON ' + self.sForm + '\n')
        #...and all the lexemes for this POS:
        for Lexeme in self.Lexemes:
            #Upper side is citation form + paradigm class, lower side is underlying form + paradigm class:
            LexiconOut.write('\t%s%s:%s%s\t#;\n' %(Lexeme.sCitationForm, Lexeme.sParadigm, Lexeme.sUnderlyingForm, Lexeme.sParadigm))
        LexiconOut.close()


#--------------------------Class RawParadigmClass---------------------------
class RawParadigmClass:
    "Keeps track of a raw ParadigmClass, its normalized form, and its count in the input."

    def __init__(self, sRawParadigmClass, sNormalizedForm):
        self.RawParadigmClass = sRawParadigmClass
            #Redundant, since this object is stored in a dictionary indexed by sRawParadigmClass
        self.NormalizedForm = sNormalizedForm
        self.count = 1

    def increment(self):
        "Increment my count."
        self.count = self.count + 1


#-------------------------------------Main-----------------------------
#--------------Initialization:
(LexiconIn, XfstOut, ErrOut) = OpenInputOutput()

ParadigmClasses = {} #Used to convert non-normalized Paradigm Classes to their
                     # normalized form; indexed by raw paradigm classes
RawPos2POS = {}     #Used to convert non-normalized POSs (which may contain punctuation)
                    # to their normalized form (no punctuation except underscore), and
                    # to hold a count of non-normalized POSs; indexed by raw POSs,
                    # value is a POS instance
POSs = {}           #Holds a set of POS instances, one for each pre-defined legit POS.
#Add all the legit POSs to the latter dictionary:
for sPOS in sPOSs:
    POSs[sPOS] = POS(sPOS)
#...and set up a special list of uninflected forms:
UninflectedForms = []

#Read in everything, so we can output the records by POSs:
RecordLines = LexiconIn.readlines()
LexiconIn.close()

#Comment header in xfst file:
XfstOut.write('!Generated from %s by NahTab2MultLexc.py on %s.\n'
              %(LexiconIn.name, time.strftime('%d %B %Y at time %H:%M', time.localtime()))
             )

#--------------Process records:
#Get a list of valid records, and normalize them:
iRecordLine = 0
for RecordLine in RecordLines:
    iRecordLine = iRecordLine + 1
    if RecordLine[0] != '!' and not re.match('\s*\n', RecordLine):
        #Ignore comment lines and blank lines
        try:
            Record = Lexeme(RecordLine)
        except (Lexeme.BackslashWarning, Lexeme.FieldParseWarning, Lexeme.EmptyStemWarning,
                Lexeme.StemCharsWarning, Lexeme.POSSpaceCharWarning, Lexeme.UnknownPOS,
                Lexeme.ParadigmClassSpaceCharWarning, Lexeme.POSNotSet, Lexeme.BadParadigmClass):
            pass  #Any error msg has already been output, so just go on to the next line
        if iRecordLine % 10 == 0:
            #Update the progress msg every ten entries:
            sys.stderr.write("\rNahTab2MultLexc.py: Processing record %i" %iRecordLine)
sys.stderr.write("\rNahTab2MultLexc.py: Processed %i records.\n" %iRecordLine)

#Now output the records in lexc files, one file for each POS that takes inflection:
sys.stderr.write("\nNahTab2MultLexc.py: Creating %i lexc files\n" %len(sPOSs))
for sPOS in sInflectedCategories:
    POS = POSs[sPOS]
#    sys.stderr.write("NahTab2MultLexc.py: Creating lexc file for POS %s\n" %POS) #For debugging
    POS.WriteXFSTLoaderCode(XfstOut)
    POS.WriteLexcFile()

#Also write out one lexc file for all the POSs that don't get inflected:
(UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')
LexiconOut = UTF8Writer(open('UninflectedForms.lexc', 'w'))
#Header for this lexc file:
LexiconOut.write('!Lexicon of uninflected forms generated on %s by NahTab2MultLexc.py.\n\n'
                 %(time.strftime('%d %B %Y at time %H:%M', time.localtime()))
                )
LexiconOut.write('LEXICON Root\n\tUninflectedForms ;\n\nLEXICON UninflectedForms\n')
#...and all the lexemes that are uninflected:
for Lexeme in UninflectedForms:
    #Upper side is citation form + paradigm class, lower side is underlying form + paradigm class:
    LexiconOut.write('\t%s%s:%s%s\t#;\n' %(Lexeme.sCitationForm, Lexeme.sParadigm, Lexeme.sUnderlyingForm, Lexeme.sParadigm))
LexiconOut.close()

XfstOut.close()
ErrOut.close()

